"""
Graphic user interface.
"""
from Tkinter import Tk, Frame, Menu, Text, Button, DISABLED, NORMAL, END, \
Scrollbar, E, W, N, S, NORMAL, DISABLED, StringVar, BooleanVar
import tkFileDialog
import tkMessageBox
import os
import sys
from conf import Config
if Config.visual_loaded:
    from visual import display
from solver.handlers import load_goal, solve
from solver.classes import Goal

class SomabotGUI(Frame):
    """Somabot's graphic user interface"""
    def __init__(self, master = None):
        Frame.__init__(self, master)
        self.grid()
        
        # Buttons
        self.buttons_frame = None
        self.reset_text_button = None
        self.search_button = None
        
        # Text zone
        self.text_area = None
        self.yscrollbar = None
        
        # Menu
        self.menu = None
        
        # 4 configuration variables used for radio buttons
        self.print_solution = StringVar()
        self.verbose = BooleanVar()
        self.visual = BooleanVar()
                
        self.create_widgets()

        # Result
        # The distinction between the two following attributes is for
        # upcoming changes
        self.solution = None 
        self.pieces = None

        # 3D scene
        self.display = None

    def create_widgets(self):
        """Call the widgets' creation methods"""
        # Buttons
        self.create_buttons_frame()
        self.create_reset_text_button()
        self.create_search_button()

        # Text zone
        self.create_text_area()

        # Menus
        self.menu = Menu(self.master)
        self.create_file_menu()
        self.create_option_menu()
        self.create_help_menu()
        self.master.config(menu = self.menu)

    # Start widgets creation

    def create_text_area(self):
        """The main text zone, used for stdout and stderr and its scrollbar"""
        self.yscrollbar = Scrollbar(self)
        self.yscrollbar.grid(row=0, column=1, sticky=N+S)
        self.textarea = Text(
            self,
            width = 80,
            height = 15,
            yscrollcommand = self.yscrollbar.set
        )
        self.yscrollbar.config(command = self.textarea.yview)
        self.textarea.grid(row = 0, column = 0)
        self.textarea.write = func_write
        self.textarea.config(state = DISABLED)

    def create_buttons_frame(self):
        """Create the buttons' container"""
        self.buttons_frame = Frame(self)
        self.buttons_frame.grid(row = 1, column = 0)

    def create_reset_text_button(self):
        """Create and place the reset text button"""
        self.reset_text_button = Button(self.buttons_frame)
        self.reset_text_button['text'] = 'Clear log'
        self.reset_text_button['command'] = self.reset_text
        self.reset_text_button.grid(row = 0, column = 0, padx = 2, pady = 5)

    def create_search_button(self):
        """Create and place the search button"""
        self.search_button = Button(self.buttons_frame)
        self.search_button['text'] = 'Search for a solution'
        self.search_button['command'] = self.search_solution
        self.search_button.grid(row = 0, column=1, padx = 2, pady = 5)
        self.search_button.config(state = DISABLED)

    def create_file_menu(self):
        """Create and place the file menu and its items"""
        file_menu = Menu(self.menu, tearoff = 0)
        self.menu.add_cascade(label = 'File',menu=file_menu)
        file_menu.add_command(
            label = 'Open CSV shape file...',
            command = self.load_goal
        )
        file_menu.add_command(label = 'Quit', command = self.close)

    def create_help_menu(self):
        """Create and place the help menu and its items"""
        help_menu = Menu(self.menu, tearoff = 0)
        self.menu.add_cascade(label = 'Help', menu = help_menu)
        help_menu.add_command(label = 'About', command = self.show_about)
        help_menu.add_command(label = 'License', command = self.show_license)

    def create_option_menu(self):
        """Create and place the option menu and its items"""
        option_menu = Menu(self.menu, tearoff = 0)
        self.menu.add_cascade(label = 'Options', menu = option_menu)

        # Result output
        print_result_menu = Menu(option_menu, tearoff = 0)
        self.print_solution.set(Config.print_solution)
        print_result_menu.add_radiobutton(
            label = "Pieces 'coordinates",
            command = self.set_print_solution,
            variable = self.print_solution,
            value = 'coordinates'
        )
        print_result_menu.add_radiobutton(
            label = 'Existence',
            command = self.set_print_solution,
            variable = self.print_solution,
            value = 'has'
        )
        option_menu.add_cascade(
            label = 'Print solution',
            menu = print_result_menu
        )

        # 3D rendering
        visual_menu = Menu(option_menu, tearoff = 0)
        
        if Config.visual_loaded:
            self.visual.set(Config.visual)
            visual_menu.add_radiobutton(
                label = 'Yes',
                command = self.set_visual,
                variable = self.visual,
                value = True
            )
            visual_menu.add_radiobutton(
                label = 'No',
                command = self.set_visual,
                variable = self.visual,
                value = False
            )
            state = NORMAL

        else:
            state = DISABLED
        
        option_menu.add_cascade(
            label = '3D Rendering',
            menu = visual_menu,
            state = state
        )

        # Verbosity
        self.verbose.set(Config.verbose)
        verbosity_menu = Menu(option_menu, tearoff = 0)
        verbosity_menu.add_radiobutton(
            label = 'Yes',
            command = self.set_verbose,
            variable = self.verbose,
            value = True
        )
        verbosity_menu.add_radiobutton(
            label = 'No',
            command = self.set_verbose,
            variable = self.verbose,
            value = False
        )
        option_menu.add_cascade(label = 'Verbosity', menu = verbosity_menu)
        
    # End widgets creation

    # Start commands

    def reset_text(self):
        """Clear the text area"""
        self.textarea.config(state = NORMAL)
        self.textarea.delete(1.0, END)
        self.textarea.config(state = DISABLED)

    def search_solution(self):
        """Search a solution button's event handler"""

        if self.display and Config.visual:
            # Got to hide the scene because it makes the search twice longer
            self.display.visible = False

            if Config.verbose:
                sys.stdout.write("""\
The 3D visualisation window will come back at the end of the search.
""")
        # Case of search on the same shape, pieces must be hidden
        if self.pieces and Config.visual_loaded and Config.visual:
            
            for piece in self.pieces:
                piece.hide()
             
        self.textarea.config(state = NORMAL)
        self.solution = solve()
        self.textarea.see(END)
        self.search_button.config(state = DISABLED)
        
        if self.solution and Config.visual_loaded and Config.visual:
            
            if self.display and Config.visual:
                self.display.visible = True
                Goal().hide()

            self.show_solution()

        elif Config.visual_loaded and Config.visual:
            self.display.visible = True

    def show_solution(self):
        """Show the next solution"""
        # Will be useful in next versions
        
        if self.pieces:
            
            for piece in self.pieces:
                piece.hide()

        self.pieces = []
            
        solution = self.solution

        for piece in solution:
            piece.show()
            self.pieces.append(piece)

    def load_goal(self):
        """Choose a CSV file and load the goal from it"""
        file_name = tkFileDialog.askopenfilename(
            initialdir = 'examples' + os.sep + 'possible',
            parent = self.master,
            title = 'Open a CSV shape file...',
        )

        if file_name != '':

            # Display initialization if needed
            if Config.visual_loaded and Config.visual and not self.display:
                self.display = display()
                self.display.exit = False
                self.display.select()

            # Got to hide everything because Visual Python doesn't
            # handle deletion.
            if Config.visual_loaded and Config.visual and Goal():
                Goal().hide()

            if Config.visual_loaded and Config.visual and self.pieces:
                
                for piece in self.pieces:
                    piece.hide()
            
            load_goal(file_name)
            self.textarea.config(state = NORMAL)
            self.textarea.see(END)
            self.textarea.config(state = DISABLED)
            
            if Goal():
                self.search_button.config(state = NORMAL)
                self.search_button['text'] = 'Search for a solution'
                self.search_button['command'] = self.search_solution
                self.current_solution_index = 0
                self.solution = None
                
            if Config.visual_loaded and Config.visual and Goal():
                self.display.visible = True
                Goal().show()
                
                if Config.verbose:
                    sys.stdout.write(
"""Right click on the scene and move the cursor to rotate the scene.
Left click + Right click on the scene and move the cursor to zoom in and out.

""")

    def show_license(self):
        """Show program's license"""
        f = open('data' + os.sep + 'license.txt')
        tkMessageBox.showinfo('License', f.read())

    def show_about(self):
        """Show version's number"""
        tkMessageBox.showinfo('About', 'Somabot, version %.1f ' \
                              % Config.version)

    def set_print_solution(self):
        """Set the result output"""
        Config.print_solution = self.print_solution.get()

    def set_verbose(self):
        """Enable/disable verbosity"""
        Config.verbose = self.verbose.get()

    def set_visual(self):
        """Enable/disable 3D rendering"""
        Config.visual = self.visual.get()
        if Config.visual:
            
            if self.display:
                self.display = display()
                self.display.visible = True
                self.display.select()
                self.display.exit = False
            
        else:
            
            if self.display:
                self.display.visible = False
                self.display = None

    def close(self):
        """Close main window and visual's window"""
        if self.display:
            self.display.visible = False
            
        self.master.quit()

    # End commands

app = None

def start_loop():
    """Init the GUI"""
    global app
    root = Tk()
    root.title('Somabot')
    app = SomabotGUI(master=root)
    sys.stdout = app.textarea
    sys.stderr = app.textarea
    sys.stdout.write('Welcome!\n')

    if Config.visual_loaded:
        sys.stdout.write('Visual Python loaded, 3D rendering enabled.\n\n')
    
    else:
        sys.stdout.write('Visual Python missing, 3D rendering disabled.\n\n')
        
    app.mainloop()
    root.destroy()

def func_write(s):
    """Used as stdout and stderr' write function"""
    global app
    app.textarea.config(state = NORMAL)
    app.textarea.insert(END, s)
    app.textarea.config(state = DISABLED)
