"""
Modelization of the Soma's cube entities.
Use cartesian coordinates for input and output, but the processings are
performed from the bindings between cubes.
"""
try:
    from visual import box, color

except ImportError:
    pass

from random import random
from copy import copy, deepcopy
from csv import reader, Sniffer
from math import fabs
from solver.input_exceptions import *

class Cube:
    """A cube and its bindings, used to build the shapes of the Soma's Cube."""

    # 3 constants for 3D rendering.
    width = 10
    height = 10
    length = 10

    def __init__(self):
        # The position of a cube bound to this one is given by its index in the
        # list.
        # x axis : indexes 0 and 3.
        # y axis : indexes 1 and 4.
        # z axis : indexes 2 and 5.
        self.bound_cubes = [None] * 6

        # Cartesian coordinates are used for input - output.
        self.coordinates = None

    def bind(self, bound_cube, index):
        """
        Bind this cube to another, on the given axis side.
        x axis : indexes 0 and 3.
        y axis : indexes 1 and 4.
        z axis : indexes 2 and 5.
        """
        if index in range(3):
            bound_cube_index = index + 3
            
        else:
            bound_cube_index = index - 3
        self.bound_cubes[index] = bound_cube
        bound_cube.bound_cubes[bound_cube_index] = self


class CubicShape:
    """A shape made of cubes."""

    def __init__(self, cubes):
        self.base_cubes = cubes
        # Visual python Box objects list
        self.boxes = set()

    def find_determining_cubes(self, cubes):
        """
        Eliminates unnecessary Cube objects from the given list,
        to know if this CubicShape fits in another one.
        """
        determining_cubes = set()
        for cube in cubes:
            cmpt = 0
            
            for i in range(6):
                
                if cube.bound_cubes[i]:
                    cmpt = cmpt + 1
                    
            if(cmpt > 1):
                determining_cubes.add(cube)
                
        return determining_cubes


class Coordinates:
    """Cartesian coordinates."""

    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z
        self.positions = set()
        # Used when passing from a cartesian representation to the inner model.
        self.bound_cubes_coordinates = [None] * 6


class ShapesFactory(object):
    """
    Builds a CubicShape from a CSV file containing a cartesian coordinates list.
    """

    # Singleton
    instance = None

    def __new__(cls, shape_type, csv_file_path):
        
        if not ShapesFactory.instance:
            ShapesFactory.instance = ShapesFactory.__ShapesFactory()
            
        return ShapesFactory.instance.build_shape(shape_type, csv_file_path)

    class __ShapesFactory:

        def __init__(self):
            # Used when passing from the cartesian model to the inner model.
            self.cubes_coordinates = set()
            # The shape's cubes.
            self.cubes = set()

        def build_shape(self, shape_type, csv_file_path):
            # Singleton's reinitialization.
            self.cubes_coordinates = set()
            self.cubes = set()
            # Coordinates' extraction from CSV file.
            csv_file = open(csv_file_path)
            csv_reader = reader(csv_file)
            has_header = Sniffer().has_header(csv_file.read(1024))
            csv_file.seek(0)
            i = 0
            
            for row in csv_reader:
                
                if i == 0 and has_header:
                    continue
                
                if len(row) != 3:
                    raise CSVShapeFileFormatException()
                
                try:
                    x, y, z = int(row[0]), int(row[1]), int(row[2])
                    
                except ValueError:
                    raise CSVShapeFileFormatException()

                if not x in range(10) or not y in range(10) \
                   or not z in range(10):
                    raise CoordinateValueException((x, y, z))
                
                self.cubes_coordinates.add(Coordinates(x, y, z))
                i += 1
                
            if shape_type == 'goal' and len(self.cubes_coordinates) != 27:
                raise CubesCountException(len(self.cubes_coordinates))
            
            # Cubes' and their bindings creation.
            # First, bindings between the coordinates are added.
            for cube_coordinates in self.cubes_coordinates:
                self.bind_cubes_coordinates(cube_coordinates,
                                             self.cubes_coordinates)
                cube = Cube()
                cube.coordinates = cube_coordinates
                cube_coordinates.cube = cube
                self.cubes.add(cube)
                
            # Then, bindings between the cubes are added.
            for cube in self.cubes:
                
                for i in range(6):
                    
                    if cube.coordinates.bound_cubes_coordinates[i]:
                        cube.bind(cube.coordinates\
                                  .bound_cubes_coordinates[i].cube, i)
    
            if shape_type == 'piece':
                
                for cube in self.cubes:
                    cube.coordinates = None
                    
                return Piece(self.cubes)
            
            if shape_type == 'goal':
                return Goal(self.cubes)

        def bind_cubes_coordinates(self, cube_coordinates, cubes_coordinates):
            """Creates bindings between the Coordinates objects."""
            # Remove coordinates that are already bound.
            other_cubes_coordinates = copy(cubes_coordinates)
            other_cubes_coordinates.remove(cube_coordinates)
            
            for bound_cube_coordinates in cube_coordinates.\
                bound_cubes_coordinates:
                
                if bound_cube_coordinates in other_cubes_coordinates:
                    other_cubes_coordinates.remove(bound_cube_coordinates)
                    
            # Bindings' creation
            # A binding is created when :
            # + two of the coordinates' components are equal
            # + the thirds differ by 1.
            x, y, z = cube_coordinates.x, cube_coordinates.y, cube_coordinates.z

            for cube_coordinates_2 in other_cubes_coordinates:
                x_2, y_2, z_2 = cube_coordinates_2.x, cube_coordinates_2.y, \
                                cube_coordinates_2.z
                x_diff, y_diff, z_diff = x - x_2, y - y_2, z - z_2
                x_bound = fabs(x_diff) == 1 and y_diff == 0 and z_diff == 0
                y_bound = fabs(y_diff) == 1 and x_diff == 0 and z_diff == 0
                z_bound = fabs(z_diff) == 1 and x_diff == 0 and y_diff == 0
                
                if(x_bound):
                    
                    if(x_diff == 1): i, j = 0, 3
                    
                    else: i, j = 3, 0
                    
                if(y_bound):
                    
                    if(y_diff == 1): i, j = 1, 4
                    
                    else: i, j = 4, 1
                    
                if(z_bound):
                    
                    if(z_diff == 1): i, j = 2, 5
                    
                    else: i, j = 5, 2
                    
                if(x_bound or y_bound or z_bound):
                    
                    cube_coordinates.bound_cubes_coordinates[i] = \
                                                            cube_coordinates_2
                    cube_coordinates_2.bound_cubes_coordinates[j] = \
                                                            cube_coordinates


class Goal(object, CubicShape):
    """The shape to create from the 7 pieces."""

    # Singleton
    instance = None

    def __new__(cls, cubes = []):
        
        if not Goal.instance:
            if cubes == []:
                return None
            Goal.instance = Goal.__Goal(cubes)
            
        return Goal.instance

    def __getattr__(self, name):
        
        if name == 'determining_cubes':
            return Goal.instance.determining_cubes
        
        return None

    @staticmethod
    def reset():
        Goal.instance = None

    class __Goal(CubicShape):
    
        def __init__(self, cubes):
            CubicShape.__init__(self, cubes)
            self.determining_cubes = self.find_determining_cubes(
                                          self.base_cubes)
            self.color = (0.5, 0.5, 0.5)
            self.opacity = 0.5

        def show(self):
            """3D rendering - Display this shape."""
            
            for b in self.boxes:
                b.visible = True
                
            if self.boxes:
                return
            
            for cube in self.base_cubes:
                coordinates = cube.coordinates
                self.boxes.add(box(pos = (coordinates.x * Cube.width,
                                             coordinates.y * Cube.height,
                                             coordinates.z * Cube.length),
                                   length = Cube.length,
                                   height = Cube.height,
                                   width = Cube.width,
                                   color = self.color,
                                   opacity = self.opacity))

        def hide(self):
            """3D rendering - Hide this shape."""
            
            for b in self.boxes:
                b.visible = False


class Rotation:
    """An hashable wrapper for piece's rotations cubes list."""

    def __init__(self, cubes):
        self.cubes = cubes


class Piece(CubicShape, object):
    """The 6 tetracubes and the tricube."""

    def __init__(self, cubes):
        CubicShape.__init__(self, cubes)
        self.name = ''
        self.rotations = set()
        self.find_rotations()
        # Possible positions in the goal
        self.positions = set()
        self.current_position = None
        # Used for 3d rendering
        self.color = color.blue
        self.opacity = 0.9
        self.rotation_boxes = set()

    def __str__(self):
        res = self.name + ':\n'
        
        if not self.current_position:
            res += 'No position.\n'
            
        else:
            res += str(self.current_position)
            
        return res

    def find_positions(self):
        """Find all possible positions in the goal, ignoring other pieces."""
        goal_cubes = Goal().determining_cubes
        pos_i = 1

        for rotation in self.rotations:
            
            # Iterates over goal's cubes looking for the rotation's pattern.
            for goal_cube in goal_cubes:
                coordinates = [goal_cube.coordinates]
                match = True
                it = iter(rotation.cubes)
                cube = it.next()
                
                for i in range(len(rotation.cubes)):
                    
                    for j in range(6):
                        
                        if cube.bound_cubes[j] and \
                           not goal_cube.bound_cubes[j]:
                            match = False
                            break
                        
                        elif cube.bound_cubes[j] and \
                             goal_cube.bound_cubes[j]:

                            if not goal_cube.bound_cubes[j].coordinates\
                               in coordinates:
                                coordinates.append(
                                    goal_cube.bound_cubes[j].coordinates
                                    )
                            
                    if match and i < (len(rotation.cubes) - 1):
                        # Continues with the next determining cube
                        cube_2 = it.next()
                        index = cube.bound_cubes.index(cube_2)
                        goal_cube = goal_cube.bound_cubes[index]
                        cube = cube_2
                        
                    else:
                        break
                    
                if match:
                    position = Position(self, coordinates)
                    position.name = '%s%i' % (self.name[0:1], pos_i)
                    pos_i += 1
                    
                    for c in coordinates:
                        c.positions.add(position)
                    self.positions.add(position)

    def find_rotations(self):
        """
        Find the distinct rotations of a piece.
        Considers the 24 possible rotations, and save the unique ones.
        """
        
        # Puts successively the 6 faces of the piece at the same point of view.
        for i in range(6):
            cubes = deepcopy(self.base_cubes)
            
            if i == 0:
                pass
            
            elif i in range(1, 4):
                
                for k in range(i):
                    cubes = self.rotate(cubes, 0)
                    
            elif i in range(4, 6):
                cubes = self.rotate(cubes, 2)
                if i == 5:
                    cubes = self.rotate(cubes, 2)
                    
            if not self.has_rotation(cubes):
                # Save the position.
                self.save_rotation(cubes)

            # Then rotates 3 times on the same axis and save the positions.
            for j in range(3):
                _cubes = deepcopy(cubes)
                
                for k in range(j + 1):
                    _cubes = self.rotate(_cubes, 1)
                    
                if not self.has_rotation(_cubes):
                    self.save_rotation(_cubes)
            
    def rotate(self, cubes, axis_index):
        """Rotate the cubes' shape by 90 degrees on an axis."""
        # 3 axis : x, y, z
        axes = [[0, 2], [0, 1], [1, 2]]
        axis = axes[axis_index]
        
        for cube in cubes:
            c1 = cube.bound_cubes[axis[0]]
            c2 = cube.bound_cubes[axis[0] + 3]
            cube.bound_cubes[axis[0]] = None
            cube.bound_cubes[axis[0] + 3] = None
            c3 = cube.bound_cubes[axis[1]]
            c4 = cube.bound_cubes[axis[1] + 3]
            cube.bound_cubes[axis[1]] = None
            cube.bound_cubes[axis[1] + 3] = None
            cube.bound_cubes[axis[0]] = c3
            cube.bound_cubes[axis[0] + 3] = c4
            cube.bound_cubes[axis[1]] = c2
            cube.bound_cubes[axis[1] + 3] = c1
            
        return cubes

    def has_rotation(self, subject):
        """Is the given Cube list already registered ?"""
        
        for rotation in self.rotations:
            s_cubes = self.find_determining_cubes(subject)
            
            for s_cube in s_cubes:
                it = iter(rotation.cubes)
                cube = it.next()
                match = True
                
                for i in range(len(rotation.cubes)):
                    
                    for j in range(6):
                        
                        if (cube.bound_cubes[j] and \
                           not s_cube.bound_cubes[j]) \
                           or \
                           (not cube.bound_cubes[j] and \
                           s_cube.bound_cubes[j]):
                            match = False
                            break
                        
                    if match and i < (len(rotation.cubes) - 1):
                        cube_2 = it.next()
                        index = cube.bound_cubes.index(cube_2)
                        cube = cube_2
                        s_cube = s_cube.bound_cubes[index]
                        
                    else:
                        break
                    
            if match:
                return True
            
        return False

    def save_rotation(self, cubes):
        """Saves the rotation's determining cubes to the rotations list."""
        cubes = self.find_determining_cubes(cubes)
        self.rotations.add(Rotation(cubes))

    def show(self, position = None):
        """
        3D rendering - Display a position of this piece.
        If the piece is placed and no parameter is given,
        shows the piece's current position.
        """
        self.hide()
        self.boxes = set()
        
        if not position and self.current_position:
            position = self.current_position
            
        for coordinates in position.coordinates:
            self.boxes.add(box(pos = (coordinates.x * Cube.width,
                                         coordinates.y * Cube.height,
                                         coordinates.z * Cube.length),
                               length = Cube.length,
                               height = Cube.height,
                               width = Cube.width,
                               color = self.color,
                               opacity = self.opacity))

    def hide(self):
        """3D rendering - Hide the current displayed piece's position."""
        
        for box in self.boxes:
            box.visible = False


class Position:
    """A set of coordinates corresponding to one piece in the goal."""

    def __init__(self, piece, coordinates):
        self.coordinates = coordinates
        self.piece = piece
        self.name = ''
        # Incompatible other pieces' positions.
        self.incompatible_positions = set()
        self.compatible_positions_by_pieces = None
        self.locked_at_step = None
        self.chosen_at_step = None

    def __str__(self):
        res = ''
        
        for coordinates in self.coordinates:
            res += '(%i,%i,%i)\n' % \
                   (coordinates.x, coordinates.y, coordinates.z)
             
        return res
