"""
Handle the successive steps of a search for a solution to a Soma's cube puzzle.
"""
try:
    from cPickle import load

except ImportError:
    from pickle import load

from copy import copy
from random import randrange, shuffle
from time import clock
import sys
import os
from conf import Config
from solver.classes import ShapesFactory, Goal, Cube
from solver.input_exceptions import *
from visual import box


def search_solution(pieces, all_solutions = False):
    """
    Searches a solution once that the goal, the pieces and their positions'
    have been initialized.
    """
    start_time = clock()

    if all_solutions:
        solutions = []

    n0 = pieces[0].name; n1 = pieces[1].name; n2 = pieces[2].name
    n3 = pieces[3].name; n4 = pieces[4].name; n5 = pieces[5].name
    n6 = pieces[6].name

    for pos0 in pieces[0].positions:
        
        # Positions from piece 1 compatible with the current piece 0' position.
        compatible_1 = pos0.compatible_positions_by_pieces[n1]

        for pos1 in compatible_1:

            # Positions from piece 2 compatible with the current piece 0'
            # position and the current piece 1' position.
            compatible_2 = pos0.compatible_positions_by_pieces[n2] \
            & pos1.compatible_positions_by_pieces[n2]

            # Piece 2
            # Iterate over all positions from piece 2 that are compatible with 
            # the chosen position from piece 1 and the chosen position from
            # piece 0.
            for pos2 in compatible_2:

                compatible_3 = pos0.compatible_positions_by_pieces[n3] \
                & pos1.compatible_positions_by_pieces[n3] \
                & pos2.compatible_positions_by_pieces[n3]

                # Piece #3
                for pos3 in compatible_3:

                    compatible_4 = pos0.compatible_positions_by_pieces[n4] \
                    & pos1.compatible_positions_by_pieces[n4] \
                    & pos2.compatible_positions_by_pieces[n4] \
                    & pos3.compatible_positions_by_pieces[n4]

                    # Piece #4
                    for pos4 in compatible_4:
                    
                        compatible_5 = pos0.compatible_positions_by_pieces[n5] \
                        & pos1.compatible_positions_by_pieces[n5] \
                        & pos2.compatible_positions_by_pieces[n5] \
                        & pos3.compatible_positions_by_pieces[n5] \
                        & pos4.compatible_positions_by_pieces[n5]

                        # Piece #5
                        for pos5 in compatible_5:
                            
                            compatible_6 = \
                            pos0.compatible_positions_by_pieces[n6] \
                            & pos1.compatible_positions_by_pieces[n6] \
                            & pos2.compatible_positions_by_pieces[n6] \
                            & pos3.compatible_positions_by_pieces[n6] \
                            & pos4.compatible_positions_by_pieces[n6] \
                            & pos5.compatible_positions_by_pieces[n6]

                            if compatible_6:

                                if all_solutions:
                                    solutions.append(
                                        [pos0, pos1, pos2, pos3,
                                         pos4, pos5, compatible_6.pop()]
                                    )

                                else:

                                    if Config.verbose:
                                        sys.stdout.write('Found a solution in \
%.2f seconds.\n' % (clock() - start_time))

                                    positions = [pos0, pos1, pos2, pos3,
                                         pos4, pos5, compatible_6.pop()]

                                    for pos in positions:
                                        pos.piece.current_position = pos

                                    return pieces

    if not all_solutions:

        if Config.verbose:
            sys.stdout.write('Found no solution in %.2f seconds.\n' % \
                  (clock() - start_time))

        return None

    if Config.verbose:
        
        if solutions:
            sys.stdout.write('Found %i solution(s) in %.2f seconds.\n' % \
                  (len(solutions), (clock() - start_time)))

        else:
            sys.stdout.write('Found no solution in %.2f seconds.\n' % \
                  (clock() - start_time))

    return solutions

def solve(all_solutions = False):
    """
    All in one, must be called after the goal's loading. Load pieces,
    find positions and search for a solution.
    Return the pieces if a solution exists, None otherwise.
    """
    t = clock()
    pieces = load_pieces()
    pieces = init_positions(pieces)
    
    if not pieces:
        return None

    if Config.verbose:
        sys.stdout.write('\n')

    pieces = search_solution(pieces, all_solutions)
    
    if pieces:

        sys.stdout.write('At least one solution.\n\n')

        if Config.print_solution == 'coordinates':

            for piece in pieces:
                sys.stdout.write(str(piece))

        if Config.print_solution == 'visual':

            for piece in pieces:
                piece.show()            

    else:
        sys.stdout.write('No solution.\n')

    if Config.verbose:
        sys.stdout.write('\nTotal elapsed time: %.2f seconds.\n' \
                         % (clock() - t))
        
    return pieces

def load_goal(goal_csv_file_path):
    """
    Returns True if a shape was loaded from the given path, None otherwise.
    As Goal is a singleton, it can be directly accessed.
    """
    Goal.reset()
    error = True
    
    try:
        goal = ShapesFactory('goal', goal_csv_file_path)
        error = False
        
    except IOError:
        msg = "Couldn't open this file.\n"
        
    except CubesCountException as exc:
        msg = str(exc)
        
    except CSVShapeFileFormatException as exc:
        msg = str(exc)

    except CoordinateValueException as exc:
        msg = str(exc)
    
    if error:
        sys.stderr.write('Error loading: \'' + goal_csv_file_path + '\'\n')
        sys.stderr.write(msg)
        return False
    
    if Config.verbose:
        msg = 'Goal loaded: \'%s\'\n' % goal_csv_file_path
        sys.stdout.write(msg)
        
    return True

def load_pieces():
    """Load pieces from their pickle file."""
    pieces_file = 'data' + os.sep + 'pieces' + os.sep + 'pieces'
    f_in = open(pieces_file, 'rb')
    pieces = load(f_in)
    
    if Config.verbose:
        msg = 'Pieces loaded.\n'
        sys.stdout.write(msg)
        
    return pieces

def init_positions(pieces):
    """
    Find all the pieces' positions and their compatibilies.
    Returns the pieces if each of them has at least one position,
    None otherwise,
    """
    t = clock()

    if Config.verbose:
        sys.stdout.write('\n')
    
    for piece in pieces:
        # Goal is implemented as a singleton and it's called in
        # Piece.find_positions
        piece.find_positions()
        
        if Config.verbose:
            msg = 'Found %i positions for %s.\n' % (len(piece.positions),
                                                  piece.name)
            sys.stdout.write(msg)

    # Incompatibilites between positions.
    for piece in pieces:
        
        if len(piece.positions) == 0:
            msg = '\nNo solution.\n'
            sys.stdout.write(msg)
                
            return None

    for cube in Goal().base_cubes:
        
        for position in cube.coordinates.positions:
            position.incompatible_positions |= cube.coordinates.positions

    for cube in Goal().base_cubes:
        
        cube.coordinates.positions = set()

    for piece in pieces:

        for position in piece.positions:
            incompatible_positions_copy = position.incompatible_positions.copy()
            
            for incompatible_position in position.incompatible_positions:
                
                if incompatible_position.piece == piece:
                    incompatible_positions_copy.remove(incompatible_position)
                    
            position.incompatible_positions = incompatible_positions_copy

    # Compatibilites between positions
    for piece in pieces:

        for position in piece.positions:
            c_pos = []
            
            for piece_2 in pieces:

                if piece_2 == piece:
                    continue
                
                c_pos.append(piece_2.positions - \
                             position.incompatible_positions)

            # Building dictionnaries keys
            el = c_pos[0].pop(); n0 = el.piece.name; c_pos[0].add(el)
            el = c_pos[1].pop(); n1 = el.piece.name; c_pos[1].add(el)
            el = c_pos[2].pop(); n2 = el.piece.name; c_pos[2].add(el)
            el = c_pos[3].pop(); n3 = el.piece.name; c_pos[3].add(el)
            el = c_pos[4].pop(); n4 = el.piece.name; c_pos[4].add(el)
            el = c_pos[5].pop(); n5 = el.piece.name; c_pos[5].add(el)
            
            position.compatible_positions_by_pieces = {
                n0 : c_pos[0],
                n1 : c_pos[1],
                n2 : c_pos[2],
                n3 : c_pos[3],
                n4 : c_pos[4],
                n5 : c_pos[5]
            }

    if Config.verbose:
        msg = 'Positions and their incompatibilites initialized in %.2f \
seconds.\n' % (clock() - t)
        sys.stdout.write(msg)
        
    return pieces

# Start - Pieces ordering fonctions

def sort_by_ascending_positions_count(pieces):
    """Sort the pieces by positions counts, ascending order."""
    ordered_pieces = []
    pos_counts = []
    
    for piece in pieces:
        pos_counts.append(len(piece.positions))
        
    for i in range(0, 7):
        m = min(pos_counts)
        x = pos_counts.index(m)
        ordered_pieces.append(pieces[x])
        pos_counts[x] = 10e10

    return ordered_pieces

# End - Position selection functions
